import torch
import torch.nn.functional as F
import numpy as np
import matplotlib.pyplot as plt
import torchvision
from torchvision import transforms
import os
from PIL import Image
import cv2
from torchvision.datasets import ImageFolder
from torchvision.utils import save_image

from .gradCAM import GradCAM

classes = ('plane', 'car', 'bird', 'cat',
            'deer', 'dog', 'frog', 'horse', 'ship', 'truck')

pred_list= []
true_list = []
device = 'cuda'

def inv_norm(image):
    inv_norm_transform = transforms.Normalize(
        mean=(-0.4914/0.2023, -0.4822/0.1994, -0.4465/0.2010),
        std=(1/0.2023, 1/0.1994, 1/0.2010))
    return inv_norm_transform(image)
    
def _test_mis(model, device, test_loader):
    model.eval()
    correct = 0
    tloss = 0

    true_label = torch.Tensor([]).to(device)
    true_label = true_label.long()
    pred_label = torch.Tensor([]).to(device)
    pred_label = pred_label.long()
    misclass_image = torch.Tensor([]).to(device)

    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)

            output = model(data)

            tloss += F.nll_loss(output, target, reduction='sum').item()
            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()

            index_wrong = ~pred.eq(target.view_as(pred))[:,0]
            true_label = torch.cat((true_label, target.view_as(pred)[index_wrong]), dim=0)
            pred_label = torch.cat((pred_label, pred[index_wrong]), dim=0)
            misclass_image = torch.cat((misclass_image, data[index_wrong]), dim=0)

    tloss /= len(test_loader.dataset)
    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)\n'.format(
            tloss, correct, len(test_loader.dataset),
            100 * correct/len(test_loader.dataset)))
            
    return true_label, pred_label, misclass_image

def mis(model, device, test_loader, nimage = 64):
    """Display the 'nimage' number of misclassified images."""
    
    #############
    #Create directory for saving misclassified images 
    dirName = '/content/mis_class/'
    subdir = '/content/mis_class/images'

    # Create target Directory if don't exist
    if not os.path.exists(dirName) and not os.path.exists(subdir):
        os.mkdir(dirName)
        os.mkdir(subdir)
        print("Directory " , dirName ,  " Created ")
        print("Directory " , subdir ,  " Created ")
    elif not os.path.exists(subdir):
        os.mkdir(subdir)
        print("Directory " , subdir ,  " Created ")
    else:
        print("Directory " , dirName, 'and', subdir ,  " already exists")
    ###############

    tlab, plab, img= _test_mis(model, device, test_loader)

    print(img.shape)
    plt.figure(figsize=(16,16))

    for index in range(0, nimage):
        plt.subplot(int(np.sqrt(nimage)), int(np.sqrt(nimage)), index+1)
        plt.xticks([])
        plt.yticks([])
        x = inv_norm(img[index])      # unnormalize
        x = x.permute(1, 2, 0) # (C, M, N) -> (M, N, C)
        x = x.cpu().numpy()

        ###########
        #To Save mis classified image for further use in GradCAM
        path = f'/content/mis_class/images/mis_{index+10}.png'
        mis_img = Image.fromarray((x.squeeze() * 255).astype(np.uint8))
        mis_img.save(path)
        #########

        plt.imshow(x.squeeze(), cmap='gray_r', interpolation= 'bilinear')
        plt.setp(plt.title(f'Predicted: {classes[plab[index,0]]}'), color= 'red')
        pred_list.append(classes[plab[index,0]])
        plt.setp(plt.xlabel(f'Ground Truth: {classes[tlab[index,0]]}'), color= 'blue')
        true_list.append(classes[tlab[index,0]])
        plt.tight_layout()

def _visualize_cam(mask, img, hm_lay=0.5, img_lay=0.5, alpha=1.0):
    """Make heatmap from mask and synthesize GradCAM result image using heatmap and img.
    Args:
        mask (torch.tensor): mask shape of (1, 1, H, W) and each element has value in range [0, 1]
        img (torch.tensor): img shape of (1, 3, H, W) and each pixel value is in range [0, 1]
    Return:
        heatmap (torch.tensor): heatmap img shape of (3, H, W)
        result (torch.tensor): synthesized GradCAM result of same shape with heatmap.
    """
    heatmap = (255 * mask.squeeze()).type(torch.uint8).cpu().numpy()
    heatmap = cv2.applyColorMap(heatmap, cv2.COLORMAP_JET)
    heatmap = torch.from_numpy(heatmap).permute(2, 0, 1).float().div(255)
    b, g, r = heatmap.split(1)
    heatmap = torch.cat([r, g, b]) * alpha

    result = heatmap*hm_lay + img.cpu()*img_lay
    result = result.div(result.max()).squeeze()

    return heatmap, result

def gen_cam(model, layer, class_idx= None, hm_lay= 0.5, img_lay= 0.5):
    
    #############
    #Create directory for saving GradCAM images.

    # Create target Directory if don't exist
    if class_idx is None:
        dirName = '/content/result_pred/'
        dir2Name = '/content/heatmap_pred/'
        if not os.path.exists(dirName) or not os.path.exists(dirName):
            os.mkdir(dirName)
            os.mkdir(dir2Name)
            print("Directory " , dirName ,  " Created ")
            print("Directory " , dir2Name ,  " Created ")
        else:
            print("Directory " , dirName , " already exists")
            print("Directory " , dir2Name , " already exists")
    else:
        dirName = '/content/result_act/'
        dir2Name = '/content/heatmap_act/'
        if not os.path.exists(dirName) or not os.path.exists(dirName):
            os.mkdir(dirName)
            os.mkdir(dir2Name)
            print("Directory " , dirName ,  " Created ")
            print("Directory " , dir2Name ,  " Created ")
        else:
            print("Directory " , dirName , " already exists")
            print("Directory " , dir2Name , " already exists")
    
    model_layer = getattr(model, layer)
    gradcam = GradCAM(model, model_layer)

    transform = transforms.Compose([
                                    transforms.ToTensor(),
                                    transforms.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010])])
    dataset = ImageFolder(root='/content/mis_class/', transform=transform)
    dataloader = torch.utils.data.DataLoader(dataset=dataset, batch_size=1, shuffle=False)
    
    for index, batch in enumerate(dataloader):
      normed_torch_img, _ = batch

      torch_img = inv_norm(normed_torch_img[0])
      if class_idx is None:
          class_idx_ = None
      else:
          class_idx_ = classes.index(class_idx[index])
      mask, _ = gradcam(normed_torch_img.to(device), class_idx = class_idx_)
      heatmap, result = _visualize_cam(mask, torch_img, hm_lay, img_lay)
      if class_idx is None:
        fp_path_heat = f'/content/heatmap_pred/map{index+10}_{layer}.png'
        save_image(heatmap, fp=fp_path_heat)
        fp_path_res = f'/content/result_pred/result{index+10}_{layer}.png'
        save_image(result, fp=fp_path_res)
      else:
        fp_path_heat = f'/content/heatmap_act/map{index+10}_{layer}.png'
        save_image(heatmap, fp=fp_path_heat)
        fp_path_res = f'/content/result_act/result{index+10}_{layer}.png'
        save_image(result, fp=fp_path_res)

def plot_pred_cam(n,l):
  fig, axes = plt.subplots(n, l+2, figsize=((l+2)*3,n*2.5))
  fig.suptitle("Grad-CAM of Mis Classified Images with respect to Predicted(wrong) Class", fontsize=20)
  for index in range(0, n):
      # plt.subplot(n, l+1, index+1)
      plt.xticks([])
      plt.yticks([])
      axes[index, 0].text(0.5, 0.5, f'Predicted {pred_list[index]} \n Actual: {true_list[index]}', fontsize= 16)
      axes[index, 0].axis('off')
      path = '/content/mis_class/images/mis_' + str(index+10) + '.png'
      img = plt.imread(path)
      axes[index, 1].imshow(img, interpolation= 'bilinear')
      axes[index, 1].axis('off')
      for layer in range(l):
        path = f'/content/result_pred/result{index+10}_layer{layer+1}.png'
        img = plt.imread(path)
        axes[index, layer+2].imshow(img.squeeze(), cmap='gray_r', interpolation= 'bilinear')
        axes[index, layer+2].set_title(f'Later: {layer+1}')
        axes[index, layer+2].axis('off')
  plt.tight_layout()
  plt.subplots_adjust(top=0.97)
  plt.show()

def plot_act_cam(n,l):
  fig, axes = plt.subplots(n, l+2, figsize=((l+2)*3,n*2.5))
  fig.suptitle("Grad-CAM of Mis Classified Images with respect to Actual(correct) Class", fontsize=20)
  for index in range(0, n):
      # plt.subplot(n, l+1, index+1)
      plt.xticks([])
      plt.yticks([])
      axes[index, 0].text(0.5, 0.5, f'Predicted {pred_list[index]} \n Actual: {true_list[index]}', fontsize= 16)
      axes[index, 0].axis('off')
      path = '/content/mis_class/images/mis_' + str(index+10) + '.png'
      img = plt.imread(path)
      axes[index, 1].imshow(img, interpolation= 'bilinear')
      axes[index, 1].axis('off')
      for layer in range(l):
        path = f'/content/result_act/result{index+10}_layer{layer+1}.png'
        img = plt.imread(path)
        axes[index, layer+2].imshow(img.squeeze(), cmap='gray_r', interpolation= 'bilinear')
        axes[index, layer+2].set_title(f'Later: {layer+1}')
        axes[index, layer+2].axis('off')
  plt.tight_layout()
  plt.subplots_adjust(top=0.97)
  plt.show()

def plot_pred_heatmap(n,l):
  fig, axes = plt.subplots(n, l+2, figsize=((l+2)*3,n*2.5))
  fig.suptitle("Grad-CAM - Heatmap of Mis Classified Images with respect to Predicted(wrong) Class", fontsize=20)
  for index in range(0, n):
      # plt.subplot(n, l+1, index+1)
      plt.xticks([])
      plt.yticks([])
      axes[index, 0].text(0.5, 0.5, f'Predicted {pred_list[index]} \n Actual: {true_list[index]}', fontsize= 16)
      axes[index, 0].axis('off')
      path = '/content/mis_class/images/mis_' + str(index+10) + '.png'
      img = plt.imread(path)
      axes[index, 1].imshow(img, interpolation= 'bilinear')
      axes[index, 1].axis('off')
      for layer in range(l):
        path = f'/content/heatmap_pred/heatmap{index+10}_layer{layer+1}.png'
        img = plt.imread(path)
        axes[index, layer+2].imshow(img.squeeze(), cmap='gray_r', interpolation= 'bilinear')
        axes[index, layer+2].set_title(f'Later: {layer+1}')
        axes[index, layer+2].axis('off')
  plt.tight_layout()
  plt.subplots_adjust(top=0.97)
  plt.show()

def plot_act_heatmap(n,l):
  fig, axes = plt.subplots(n, l+2, figsize=((l+2)*3,n*2.5))
  fig.suptitle("Grad-CAM - Heatmap of Mis Classified Images with respect to Actual(correct) Class", fontsize=20)
  for index in range(0, n):
      # plt.subplot(n, l+1, index+1)
      plt.xticks([])
      plt.yticks([])
      axes[index, 0].text(0.5, 0.5, f'Predicted {pred_list[index]} \n Actual: {true_list[index]}', fontsize= 16)
      axes[index, 0].axis('off')
      path = '/content/mis_class/images/mis_' + str(index+10) + '.png'
      img = plt.imread(path)
      axes[index, 1].imshow(img, interpolation= 'bilinear')
      axes[index, 1].axis('off')
      for layer in range(l):
        path = f'/content/heatmap_act/heatmap{index+10}_layer{layer+1}.png'
        img = plt.imread(path)
        axes[index, layer+2].imshow(img.squeeze(), cmap='gray_r', interpolation= 'bilinear')
        axes[index, layer+2].set_title(f'Later: {layer+1}')
        axes[index, layer+2].axis('off')
  plt.tight_layout()
  plt.subplots_adjust(top=0.97)
  plt.show()


import abc

import torchvision.transforms as T
import albumentations as alb
from albumentations.pytorch import ToTensor
import numpy as np

class AlbumentationTrans:
    
    def __init__(self, transform):
        self.album_transform = transform

    def __call__(self, img):
        img = np.array(img)
        return self.album_transform(image=img)['image']

class AugmentationBase(abc.ABC):
    def build_transforms(self, train):
        return self.build_train() if train else self.build_test()

    @abc.abstractmethod
    def build_train(self):
        pass

    @abc.abstractmethod
    def build_test(self):
        pass


class MNIST_Transforms(AugmentationBase):

    def build_train(self):
        return T.Compose([
            T.ToTensor(),
            T.Normalize((0.1307,), (0.3081,))])

    def build_test(self):
        return T.Compose([
            T.ToTensor(),
            T.Normalize((0.1307,), (0.3081,))])


class CIFAR10_Transforms(AugmentationBase):

    def build_train(self):
        return T.Compose([
            T.RandomCrop(32, padding=4),
            T.RandomHorizontalFlip(),
            T.ToTensor(),
            T.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
        ])

    def build_test(self):
        return T.Compose([
            T.ToTensor(),
            T.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
        ])

class CIFAR10_AlbumTrans(AugmentationBase):

    def build_train(self):
        train_trans = alb.Compose([
            alb.Rotate((-7.0, 7.0)),
            #alb.Cutout(1,16,16, [0.4914, 0.4822, 0.4465]*255),
            alb.CoarseDropout(1, 16, 16, fill_value= [0.4914*255, 0.4822*255, 0.4465*255]),
            #alb.RandomCrop(224,224),
            alb.HorizontalFlip(),
            alb.Normalize(
                mean=[0.4914, 0.4822, 0.4465],
                std=[0.2023, 0.1994, 0.2010]
            ),
            ToTensor()
        ])
        return AlbumentationTrans(train_trans)

    def build_test(self):
        test_trans = alb.Compose([
            alb.Normalize(
                mean=[0.4914, 0.4822, 0.4465],
                std=[0.2023, 0.1994, 0.2010]
            ),
            ToTensor()
        ])
        return AlbumentationTrans(test_trans)




import torch
import torch.nn as nn
import torch.nn.functional as F

class Test:
    def __init__(self, model, device, test_loader, criterion):
        self.model = model
        self.device = device
        self.test_loader = test_loader

        self.test_loss = []
        self.test_acc = []
        self.criterion = criterion
        self.max_val = 0

    def test(self):
        self.model.eval()
        correct = 0
        tloss = 0

        with torch.no_grad():
            for data, target in self.test_loader:
                data, target = data.to(self.device), target.to(self.device)

                output = self.model(data)

                tloss += self.criterion(output, target).item()
                pred = output.argmax(dim=1, keepdim=True)
                correct += pred.eq(target.view_as(pred)).sum().item()

        tloss /= len(self.test_loader.dataset)
        self.test_loss.append(tloss)
        print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)\n'.format(
            tloss, correct, len(self.test_loader.dataset),
            100 * correct/len(self.test_loader.dataset)))

        self.test_acc.append(100 * correct/len(self.test_loader.dataset))

        if self.test_acc[-1] > self.max_val:
            self.max_val = self.test_acc[-1]
            path = '/content/classifier.pt'
            torch.save(self.model.state_dict(), path)


import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm

class Train:
    def __init__(self, model, device, train_loader, optimizer, criterion):
        
        self.model = model
        self.device = device
        self.train_loader = train_loader
        self.optimizer = optimizer

        self.train_loss = []

        self.train_acc = []
        self.train_endacc = []
        self.criterion = criterion

    def train(self, epoch):
        self.model.train()
        pbar = tqdm(self.train_loader)
        correct = 0
        processed = 0
        for batch_idx, (data, target) in enumerate(pbar):
            data, target = data.to(self.device), target.to(self.device)

            self.optimizer.zero_grad()

            y_pred = self.model(data)

            loss = self.criterion(y_pred, target)

            self.train_loss.append(loss)

            loss.backward()
            self.optimizer.step()

            pred = y_pred.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()
            processed += len(data)

            pbar.set_description(
                desc=f'Loss={loss.item()} Batch_id={batch_idx} Accuracy={100*correct/processed:0.2f}')
            self.train_acc.append(100 * correct/processed)
        self.train_endacc.append(self.train_acc[-1])


from torchvision import datasets
from torch.utils.data import DataLoader
import matplotlib.pyplot as plt
import numpy as np

from .augmentation import MNIST_Transforms, CIFAR10_Transforms, CIFAR10_AlbumTrans

class MNISTDataLoader:
    """
    It creates a data loader for test and train. It taken transformations from the 'augmentation' module
    """
    def __init__(self, model_transform, batch_size=64, data_dir= './root', shuffle=True, nworkers=4, pin_memory=True):
        self.data_dir = data_dir

        self.train_set = datasets.MNIST(
            self.data_dir,
            train=True,
            download=True,
            transform=model_transform.build_transforms(train=True)
        )

        self.test_set = datasets.MNIST(
            self.data_dir,
            train=False,
            download=True,
            transform=model_transform.build_transforms(train=False)
        )

        self.init_kwargs = {
            'shuffle': shuffle,
            'batch_size': batch_size,
            'num_workers': nworkers,
            'pin_memory': pin_memory
        }

    def get_loaders(self):
        return DataLoader(self.train_set, **self.init_kwargs), DataLoader(self.test_set, **self.init_kwargs)

class CIFAR10DataLoader:

    class_names = ['airplane', 'automobile', 'bird', 'cat',
                   'deer', 'dog', 'frog', 'horse', 'ship', 'truck']

    def __init__(self, model_transform, batch_size=64, data_dir= './root', shuffle=True, nworkers=4, pin_memory=True):
        self.data_dir = data_dir

        self.train_set = datasets.CIFAR10(
            self.data_dir,
            train=True,
            download=True,
            transform=model_transform.build_transforms(train=True)
        )

        self.test_set = datasets.CIFAR10(
            self.data_dir,
            train=False,
            download=True,
            transform=model_transform.build_transforms(train=False)
        )

        self.init_kwargs = {
            'shuffle': shuffle,
            'batch_size': batch_size,
            'num_workers': nworkers,
            'pin_memory': pin_memory
        }

    def get_loaders(self):
        return DataLoader(self.train_set, **self.init_kwargs), DataLoader(self.test_set, **self.init_kwargs)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as fh:
    required = fh.read().splitlines()

setuptools.setup(
    name="pynoob",
    version="0.0.2.0",
    author="Syed Abdul K",
    author_email="abdksyed@gmail.com",
    description="A n00b DeepLearning Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/abksyed/pynoob",
    download_url = "https://github.com/abksyed/pynoob/dist/pynoob-0.0.2.0.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=required,
    include_package_data=True
)



import torch
import matplotlib.pyplot as plt

classes = ('plane', 'car', 'bird', 'cat',
            'deer', 'dog', 'frog', 'horse', 'ship', 'truck')

def acc_loss(train_obj, test_obj):
    """Display the Train Loss and Accuracy graph. Test Loss and Accuracy graph."""
    fig, axs = plt.subplots(2,2,figsize=(15,10))
    axs[0, 0].plot(train_obj.train_loss)
    axs[0, 0].set_title("Training Loss")
    axs[1, 0].plot(train_obj.train_acc)
    axs[1, 0].set_title("Training Accuracy")
    axs[0, 1].plot(test_obj.test_loss)
    axs[0, 1].set_title("Test Loss")
    axs[1, 1].plot(test_obj.test_acc)
    axs[1, 1].set_title("Test Accuracy")

def testvtrain(train_obj, test_obj):
    """Display Test vs Train Accuracy plot"""
    plt.axes(xlabel= 'epochs', ylabel= 'Accuracy')
    plt.plot(train_obj.train_endacc)
    plt.plot(test_obj.test_acc)
    plt.title('Test vs Train Accuracy')
    plt.legend(['Train', 'Test'])

def class_acc(model,device, test_loader):            
    class_correct = list(0. for i in range(10))
    class_total = list(0. for i in range(10))

    with torch.no_grad():
        for data, target in test_loader:
            images, labels = data.to(device), target.to(device)
            outputs = model(images)
            _, predicted = torch.max(outputs, 1)
            c = (predicted == labels).squeeze()
            for i in range(labels.shape[0]):
                label = labels[i]
                class_correct[label] += c[i].item()
                class_total[label] += 1


    for i in range(10):
        print('Accuracy of %5s : %2d %%' % (
            classes[i], 100 * class_correct[i] / class_total[i]))


import torch
import torch.nn.functional as F
from torchvision import transforms
import numpy as np
import matplotlib.pyplot as plt
from torchsummary import summary
import torchvision

classes = ('plane', 'car', 'bird', 'cat',
            'deer', 'dog', 'frog', 'horse', 'ship', 'truck')

def model_summary(model):
    """Displays the Summary of the Architecture - All the methods used and the parameters used"""
    summary(model, input_size=(3, 32, 32))

def _imshow(img):
    inv_norm = transforms.Normalize(
        mean=(-0.4914/0.2023, -0.4822/0.1994, -0.4465/0.2010),
        std=(1/0.2023, 1/0.1994, 1/0.2010))
    img = inv_norm(img)      # unnormalize
    npimg = img.numpy()
    plt.figure()
    plt.imshow(np.transpose(npimg, (1, 2, 0)), interpolation= 'bilinear')

def display(train_loader, n= 64,):

    # get some random training images
    dataiter = iter(train_loader)
    images, labels = dataiter.next()
    for i in range(0,n,int(np.sqrt(n))):
        _imshow(torchvision.utils.make_grid(images[i: i+int(np.sqrt(n))]))
        # print labels
        plt.title(' '.join('%7s' % classes[j] for j in labels[i: i+int(np.sqrt(n))]), loc= 'left')